/***************************************************************************
 *   Copyright (C) 2006, IBM                                               *
 *                                                                         *
 *   Maintained by:                                                        *
 *   Eric Munson and Brad Peters                                           *
 *   munsone@us.ibm.com, bpeters@us.ibm.com                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#define TRACE_ON

#include <vector>
#include <string>
#include <iostream>
#include <sstream>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE // for getopt_long
#endif

#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <zlib.h>

#include <libvpd-1/component.h>
#include <libvpd-1/system.h>
#include <libvpd-1/lsvpd.h>
#include <libvpd-1/debug.h>
#include <libvpd-1/logger.h>
#include <libvpd-1/gatherer.h>
#include <libvpd-1/devicetreecollector.h>
#include <libvpd-1/vpddbenv.h>

using namespace lsvpd;
using namespace std;

int initializeDB( );
int addDevice( const string& udi );
int removeDevice( const string& udi );
int storeComponents( System* root, VpdDbEnv& db );
int storeComponents( Component* root, VpdDbEnv& db );
void printUsage( );
void printVersion( );
int ensureEnv( const string& env );

const string ADD( "add" );
const string REMOVE( "remove" );
const string DB_DIR( "/var/lib/lsvpd" );
const string DB_FILENAME( "db" );
const string BASE( "/sys/bus" );

bool isRoot(void);

string env = DB_DIR, file = DB_FILENAME;

int main( int argc, char** argv )
{
	char opts [] = "vhp:";
	bool done = false;
	char * envVal;
	char * err;	
	char *tmp;
	string idNode;
	int ret;
	int index = 0;
	
	struct option longOpts [] = 
	{
		{ "help", 0, 0, 'h' },
		{ "path", 1, 0, 'p' },
		{ "version", 0, 0, 'v' },
		{ 0, 0, 0, 0 }
	};
	
	while( !done )
	{
		switch( getopt_long( argc, argv, opts, longOpts, &index ) )
		{
			case 'p':
				env = optarg;
				index = env.rfind( '/' );
				file = env.substr( index + 1 );
				env = env.substr( 0, index );
				break;
				
			case 'v':
				printVersion( );
				return 0;
				
			case -1:
				done = true;
				break;
				
			case 'h':
			default:
				printUsage( );
				return 0;
		}
	}

	
	/* Test to see if running as root: */
	if (!isRoot()) {
		cout << "vpdupdater must be run as root" << endl;
		return -1;
	}
	
	Logger l;	
	l.log( "Executing lsvpd_updater", LOG_NOTICE );	
	
	envVal = getenv( "ACTION" );
	if( envVal == NULL )
	{
		// If this is called without the environment variable HAL_ACTION being set it
		// means that we are building a new database from scratch.
		l.log( "Vpdupdate: Constructing full devices database", LOG_NOTICE );	
		return initializeDB( );
	}
	else
	{
		string action = string( envVal );
		l.log( " Hotplug device detected! ", LOG_ERR );
		tmp = getenv("DEVPATH");
		
		if (action == "add") {			
			if (tmp != NULL) {				
				idNode = string(tmp);	
				l.log(string("Hotplug ADD Device Path: ") + idNode , LOG_ERR );				
				addDevice(idNode);				
				l.log( " Discovery added device ", LOG_ERR );
				l.log( " ----------------------------------- ", LOG_ERR );
				return 0;
			}
			
			l.log( " Discovery failed to add device ", LOG_ERR );
			l.log( " ----------------------------------- ", LOG_ERR );
				
		}
		else if (action == "remove") {						
			if (tmp != NULL) {
				idNode = string(tmp);	
				l.log(string("Hotplug REMOVE Device Path: ") + idNode , LOG_ERR );
				ret = removeDevice(idNode);
				if (!ret) {
					l.log( " Discovery removed device ", LOG_ERR );
					l.log( " ----------------------------------- ", LOG_ERR );				
					return 0;
				}
				else {
					l.log( " Discovery FAILED to remove device ", LOG_ERR );
					l.log( " ----------------------------------- ", LOG_ERR );				
					return ret;
				}
			}
		}
	}
	
	l.log( "vpdupdate Complete", LOG_NOTICE );	
}


bool isRoot()
{	
	return (geteuid() == 0);
}



/**
 * Method does the initial population of the vpd db, this should only
 * be done once at boot time or any time that a user wishes to start with
 * a new db.
 */
int initializeDB( )
{
	if( ensureEnv( env ) != 0 )
		return -1;
	
	string fullPath = env + "/" + file;
	DIR * pDBdir = NULL;
	
	struct stat st;
	if( stat( fullPath.c_str( ), &st ) == 0 )
	{
		/*
		 * The old file is there, it needs to be timestamped, 
		 * gzipped, and moved.
		 */
		struct dirent*  ent;
		pDBdir = opendir( env.c_str( ) );
		while( ( ent = readdir( pDBdir ) ) != NULL )
		{
			string fname = ent->d_name;
			string tag = "__" + file;
			if( fname.find( tag ) != string::npos )
			{
				unlink( fname.c_str( ) );
			}
		}
		closedir( pDBdir );
		
		ostringstream os;
		os << fullPath << "." << time( NULL );
		link( fullPath.c_str( ), os.str( ).c_str( ) );
		unlink( fullPath.c_str( ) );
		
		if( stat( os.str( ).c_str( ), &st ) == 0 )
		{
			char * buffer = new char[ st.st_size + 1 ];
			gzFile gzf = NULL;
			int tot = 0, in = 0, fd = -1;
			if( buffer == NULL )
			{
				cout << "Out of memory." << endl;
				goto ZDONE;
			}
			
			memset( buffer, 0, st.st_size + 1 );
			fd = open( os.str( ).c_str( ), O_RDONLY );
			if( fd < 0 )
			{
				delete [] buffer;
				cout << "Failed to open db file " << os.str( ) << 
					" for reading. " << endl;
				goto ZDONE;
			}
			
			// Read the original file into memory.
			while( ( in = read( fd, buffer + tot, st.st_size - tot ) ) > 0 &&
				tot < st.st_size )
				tot += in;
			close( fd );
			unlink( os.str( ).c_str( ) );
			
			os << ".gz";
			gzf = gzopen( os.str( ).c_str( ), "wb9" );
			if( gzf == NULL )
			{
				delete [] buffer;
				cout << "Error openning archive file " << os.str( ) <<
					" for writing." << endl;
				goto ZDONE; 
			}
			
			in = 0;
			tot = 0;
			while( ( in = gzwrite( gzf, buffer + tot, st.st_size - tot ) ) > 0
				&& tot < st.st_size )
				tot += in;
			
			delete [] buffer;
			
			if( gzclose( gzf ) != 0 )
			{
				cout << "Failed to write compressed database, error = '" <<
					gzerror( gzf, &in ) << "'" << endl;
				goto ZDONE;
			}
ZDONE:;
		}
	}
	
	creat( fullPath.c_str( ), S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP |
								S_IROTH );
		
	Gatherer info;	
	VpdDbEnv db( env, file, DB_CREATE );
	System* root = info.getComponentTree( );
	
	int ret = storeComponents( root, db );
	//info.diplayInheritanceTree(root);

	delete root;
	return ret;
}

/**
 * Recursively descend the component tree and store each in the db.
 */
int storeComponents( Component* root, VpdDbEnv& db )
{
	if( !db.store( root ) )
	{
		return -1;
	}
	
	vector<Component*>::const_iterator i, end;
	const vector<Component*> children = root->getLeaves( );
	for( i = children.begin( ), end = children.end( ); i != end; ++i )
	{		
		int ret = storeComponents( *i, db );
		if(  ret != 0 )
		{
			return ret;
		}
	}
	return 0;
}

int storeComponents( System* root, VpdDbEnv& db )
{

	if( !db.store( root ) )
	{
		return -1;
	}
	
	vector<Component*>::const_iterator i, end;
	const vector<Component*> children = root->getLeaves( );
	for( i = children.begin( ), end = children.end( ); i != end; ++i )
	{
		int ret = storeComponents( *i, db );
		if(  ret != 0 )
		{
			return ret;
		}
	}
	return 0;
}

/**
 * Method adds a device to the vpd db.
 */
int addDevice( const string& id )
{
	Component *added;
	Gatherer info;
	VpdDbEnv db( env, file, 0 );
	added = info.hotplugAdd(id);
	info.getComponentDetails( added );
	
	if( !db.store( added ) )
	{
		return -1;
	}
	
//	if( added->getParent( ) != string( "" ) )
//	{
//		// If we can identify the parent of this node we need to tell the
//		// parent object of this child.
//		return populateParent( added->getParent( ), added->getID( ), info, db );
//	}
	
	return 0;
}

/**
 * Method removes a device from the vpd db.  This will also delete the specified device
 * from its parent's vector of children.
 */
int removeDevice( const string& id )
{
	coutd << "Attempting removal of device at idNode: " << id << endl;
	VpdDbEnv db( env, file, 0 );
	Component me = db.fetch( id );
	Component parent = db.fetch( me.getID( ) );
	parent.removeChild( me.getID( ) );
	if( parent.getID( ) != "" && !db.store( &parent ) )
	{
		return -1;
	}
	
	if( ! db.remove( id ) )
	{
		return -1;
	}
	return 0;
}

int ensureEnv( const string& env )
{
	struct stat info;
	int ret = 0;
	
	if( stat( env.c_str( ), &info ) == 0 )
	{
		return ret;
	}
	
	int idx;
	if( ( idx = env.rfind( '/' ) ) == string::npos )
	{
		return ret;
	}
	
	if( ( ret = ensureEnv( env.substr( 0, idx ) ) ) != 0 )
	{
		return ret;
	}
	
	if( mkdir( env.c_str( ),
			S_IRUSR | S_IWUSR | S_IXUSR |
			S_IRGRP | S_IWGRP | S_IXGRP |
			S_IROTH | S_IXOTH ) != 0 )
	{
		Logger logger;
		logger.log( "Failed to create directory for vpd db.", LOG_ERR );
		return -1;
	}
	return ret;
}

void printUsage( )
{
	string prefix( DEST_DIR );
	cout << "Usage: " << prefix;
	if( prefix[ prefix.length( ) - 1 ] != '/' )
	{
		cout << "/";
	}
	cout << "sbin/vpdupdater [options]" << endl;
	cout << "options:" << endl;
	cout << " --help,      -h     Prints this message." << endl;
	cout << " --version,   -v     Prints the version of vpd tools." << endl;
	cout << " --path=PATH, -pPATH Sets the path to the vpd db to PATH" << endl;
}

void printVersion( )
{
	cout << "vpdupdate " << VPD_VERSION << endl;
}
